from typing import List, Optional

from pydantic import BaseModel, ValidationError, validator

from pyngo import drf_error_details


class NestedModel(BaseModel):
    str_field: str

    @validator("str_field")
    def must_be_bar(cls, value: str) -> str:
        special_name = "bar"
        if value != special_name:
            raise ValueError(f"Name must be: '{special_name}'!")
        else:
            return value


class MyModel(BaseModel):
    int_field: Optional[int]
    nested_field: Optional[NestedModel]
    nested_list: Optional[List[NestedModel]]

    @validator("int_field")
    def must_be_special_value(cls, value: int) -> int:
        magic_number = 42
        if value != magic_number:
            raise ValueError("Must be the magic number!")
        else:
            return value


class TestToDRFError:
    def test_with_single_flat_field(self) -> None:
        try:
            MyModel(int_field=2)
        except ValidationError as e:
            assert drf_error_details(e) == {"int_field": ["Must be the magic number!"]}

    def test_with_nested_field(self) -> None:
        try:
            MyModel.parse_obj({"int_field": 42, "nested_field": {"str_field": "foo"}})
        except ValidationError as e:
            assert drf_error_details(e) == {
                "nested_field": {"str_field": ["Name must be: 'bar'!"]}
            }

    def test_with_nested_list(self) -> None:
        try:
            MyModel.parse_obj(
                {"nested_list": [{"str_field": "bar"}, {"str_field": "foo"}]}
            )
        except ValidationError as e:
            assert drf_error_details(e) == {
                "nested_list": {"1": {"str_field": ["Name must be: 'bar'!"]}}
            }



from typing import List, Type, cast

from pydantic import BaseModel
from pydantic.fields import ModelField
from typing_extensions import Literal, TypedDict

_In = Literal["query", "header", "path", "cookie"]

ParameterDict = TypedDict(
    "ParameterDict",
    {
        "name": str,
        "in": _In,
        "description": str,
        "required": bool,
        "deprecated": bool,
        "allowEmptyValue": bool,
    },
    total=False,
)

_VALID_LOCATIONS = ("query", "header", "path", "cookie")


def openapi_params(
    model_class: Type[BaseModel],
) -> List[ParameterDict]:
    """
    Returns a list of parameters for the given model class.

    Args:
        model_class (Type[BaseModel]): The model class to get parameters for.

    Raises:
        ValueError: If the model class is not a pydantic model.
        ValueError: If the model class has a field with an invalid location.
        ValueError: If the model class has a field with allowEmptyValue set for a location other than 'query'.
        ValueError: If the model class has a field with required set to False for a path parameter.

    Returns:
        List[ParameterDict]: A list of parameters for the given model class.
    """
    parameters: List[ParameterDict] = []

    for field in model_class.__fields__.values():
        if field.is_complex():
            raise ValueError("Only simple types allowed")
        else:
            parameters.append(_pydantic_field_to_parameter(field))

    return parameters


def _pydantic_field_to_parameter(field: ModelField) -> ParameterDict:
    """
    Converts a pydantic field to an OpenAPI parameter.

    Args:
        field (ModelField): The field to convert.

    Raises:
        ValueError: If the field has an invalid location.
        ValueError: If the field has allowEmptyValue set for a location other than 'query'.
        ValueError: If the field has required set to False for a path parameter.

    Returns:
        ParameterDict: The converted field.
    """
    location = field.field_info.extra.get("location", "query")
    if location not in _VALID_LOCATIONS:
        raise ValueError(f"location must be one of: {', '.join(_VALID_LOCATIONS)}")

    required = field.required
    if location == "path" and not required:
        raise ValueError("Path parameters must be required")

    field_extra = field.field_info.extra
    deprecated = field_extra.get("deprecated", False)

    args = {
        "name": field.name,
        "in": location,
        "description": field.field_info.description or "",
        "required": required,
        "deprecated": deprecated,
    }
    allow_empty_value = field_extra.get("allowEmptyValue")
    if allow_empty_value is not None and location != "query":
        raise ValueError("allowEmptyValue only permitted for 'query' values")
    elif location == "query":
        if allow_empty_value is not None:
            args["allowEmptyValue"] = allow_empty_value
        else:
            args["allowEmptyValue"] = False

    return cast(ParameterDict, args)



from typing import Dict, Optional

import pytest
from pydantic import BaseModel, Field

from pyngo import ParameterDict, openapi_params


class TestPydanticModelToOpenapiParameters:
    def test_only_primitive_types_allowed(self) -> None:
        class Model(BaseModel):
            non_primitive: Dict[str, str]

        with pytest.raises(ValueError) as exc:
            openapi_params(Model)
        assert str(exc.value) == "Only simple types allowed"

    def test_prohibits_unknown_location(self) -> None:
        class Model(BaseModel):
            path_param: int = Field(location="foo")

        with pytest.raises(ValueError) as exc:
            openapi_params(Model)
        assert str(exc.value) == "location must be one of: query, header, path, cookie"

    def test_prohibits_optional_path_params(self) -> None:
        class Model(BaseModel):
            path_param: Optional[int] = Field(location="path")

        with pytest.raises(ValueError) as exc:
            openapi_params(Model)
        assert str(exc.value) == "Path parameters must be required"

    def test_defaults_path_params_to_be_required(self) -> None:
        class Model(BaseModel):
            path_param: int = Field(location="path")

        params = openapi_params(Model)
        assert len(params) == 1
        assert params[0]["required"]

    @pytest.mark.parametrize("param_loc", ("header", "path", "cookie"))
    def test_prohibits_allow_empty_outside_of_query(self, param_loc: str) -> None:
        class Model(BaseModel):
            param: int = Field(location=param_loc, allowEmptyValue=True)

        with pytest.raises(ValueError) as exc:
            openapi_params(Model)
        assert str(exc.value) == "allowEmptyValue only permitted for 'query' values"

    def test_allow_empty_excluded_for_non_query_params(self) -> None:
        class Model(BaseModel):
            param: int = Field(location="header")

        params = openapi_params(Model)
        assert len(params) == 1
        assert "allowEmptyValue" not in params[0]

    def test_allow_empty_value_defaults_to_true_for_query_params(self) -> None:
        class Model(BaseModel):
            param: int = Field()

        params = openapi_params(Model)
        assert len(params) == 1
        assert not params[0]["allowEmptyValue"]

    def test_deprecated_defaults_to_false(self) -> None:
        class Model(BaseModel):
            param: int = Field()

        params = openapi_params(Model)
        assert len(params) == 1
        assert not params[0]["deprecated"]

    def test_optional_fields_are_not_required(self) -> None:
        class Model(BaseModel):
            param: Optional[int] = Field()

        params = openapi_params(Model)
        assert len(params) == 1
        assert not params[0]["required"]

    def test_reading_query_params(self) -> None:
        class PathParams(BaseModel):
            document_id: int = Field(
                location="query",
                description="The document id",
                required=True,
                deprecated=True,
                allowEmptyValue=True,
            )

        expected_parameters = [
            ParameterDict(
                {
                    "in": "query",
                    "name": "document_id",
                    "description": "The document id",
                    "required": True,
                    "deprecated": True,
                    "allowEmptyValue": True,
                }
            )
        ]

        assert openapi_params(PathParams) == expected_parameters



import asyncio

import pytest


@pytest.fixture
def event_loop():
    """
    Fixture for the event loop.
    Yields:
        asyncio.AbstractEventLoop: The event loop.
    """
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()



"""Pyngo: Pydantic Package for Adding Models into a Django or Django Rest Framework Project"""

__version__ = "1.3.0"

from .errors import drf_error_details
from .openapi import ParameterDict, openapi_params
from .querydict import QueryDictModel, querydict_to_dict

__all__ = (
    "ParameterDict",
    "QueryDictModel",
    "querydict_to_dict",
    "drf_error_details",
    "openapi_params",
)



from typing import Any, Dict, Sequence

from pydantic import ValidationError


def drf_error_details(exception: ValidationError) -> Dict[str, Any]:
    """
    Convert a pydantic ValidationError into a DRF-style error response.

    Args:
        exception (ValidationError): The exception to convert.

    Returns:
        Dict[str, Any]: The error response.
    """
    drf_data: Dict[str, Any] = {}
    for error in exception.errors():
        set_nested(drf_data, error["loc"], [error["msg"]])
    return drf_data


def set_nested(data: Dict[str, Any], keys: Sequence[str], value: Any) -> None:
    """
    Set a value in a nested dictionary.

    Args:
        data (Dict[str, Any]): The dictionary to set the value in.
        keys (Sequence[str]): The keys to set the value at.
        value (Any): The value to set.

    Returns:
        None
    """
    for key in keys[:-1]:
        data = data.setdefault(str(key), {})
    data[keys[-1]] = value


def get_nested(data: Dict[str, Any], keys: Sequence[str]) -> Any:
    """
    Get a value from a nested dictionary.

    Args:
        data (Dict[str, Any]): The dictionary to get the value from.
        keys (Sequence[str]): The keys to get the value at.

    Returns:
        Any: The value.
    """
    for key in keys[:-1]:
        data = data[key]
    return data[keys[-1]]



import sys
from typing import Any, Dict, Optional, Type, TypeVar

from django.http import QueryDict
from pydantic import BaseModel
from pydantic.fields import ModelField

if sys.version_info >= (3, 8):
    from typing import get_origin
else:

    def get_origin(tp: Any) -> Optional[Any]:
        try:
            return tp.__origin__
        except AttributeError:
            return None


_QueryDictModel = TypeVar("_QueryDictModel", bound="QueryDictModel")


class QueryDictModel(BaseModel):
    """
    A model that can be initialized from a QueryDict.

    This is a base class for models that can be initialized from a QueryDict.

    The QueryDictModel class is a base class for models that can be initialized

    from a QueryDict.
    """

    @classmethod
    def parse_obj(cls: Type["_QueryDictModel"], obj: Any) -> "_QueryDictModel":
        """
        Parse a QueryDict into a model.

        Returns:
            A model that was initialized from the QueryDict.
        """
        if isinstance(obj, QueryDict):
            obj = querydict_to_dict(obj, cls)
        return super().parse_obj(obj)


def querydict_to_dict(
    query_dict: QueryDict,
    model_class: Type[BaseModel],
) -> Dict[str, Any]:
    """
    Convert a QueryDict into a dictionary.

    Args:
        query_dict (QueryDict): The QueryDict to convert.
        model_class (Type[BaseModel]): The model class to use for the conversion.

    Returns:
        Dict[str, Any]: The converted dictionary.
    """
    to_dict: Dict[str, Any] = {}
    model_fields = model_class.__fields__

    for key in query_dict.keys():
        if key in model_fields and _is_list_field(model_fields[key]):
            to_dict[key] = query_dict.getlist(key)
        else:
            to_dict[key] = query_dict.get(key)
    return to_dict


def _is_list_field(field: ModelField) -> bool:
    """
    Check if a field is a list field.

    Args:
        field (ModelField): The field to check.

    Returns:
        bool: True if the field is a list field, False otherwise.
    """
    return get_origin(field.outer_type_) == list



import os
from typing import Any, Dict, List, Union

import pytest
from django.http import QueryDict
from pydantic import BaseModel

from pyngo import QueryDictModel

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"


class Model(QueryDictModel, BaseModel):
    foo: int
    bar: List[int]


@pytest.mark.parametrize(
    ("data", "expected"),
    (
        (QueryDict("foo=12&bar=12"), Model(foo=12, bar=[12])),
        ({"foo": 44, "bar": [0, 4]}, Model(foo=44, bar=[0, 4])),
    ),
)
def test_parsing_objects(
    data: Union[QueryDict, Dict[str, Any]], expected: Model
) -> None:
    got = Model.parse_obj(data)
    assert got == expected
    assert got.foo == expected.foo
    assert got.bar == expected.bar
